import bpy
from bpy.props import *
from bpy_extras import object_utils
import bmesh
import math

from . import Util
from . import Smoother

class SmoothEdgeLoopOperator(bpy.types.Operator, object_utils.AddObjectHelper):
    """Smooth Edge Loop"""
    bl_idname = 'mesh.smooth_edge_loop'
    bl_label = 'Smooth Edge Loop'
    bl_description = 'Smooth selected EdgeLoop of (preferably convex) Mesh [select at least 1 Vertex on the Mesh]'
    bl_options = {'REGISTER', 'UNDO', 'PRESET'}

    InvertLoop = BoolProperty(name = "Invert Loop", default = False, description = "Invert loop order of vertices")
    
    AddFaces = BoolProperty(name = "Add Faces", default = True, description = "Add Faces to the resulting Mesh")
    AddEdges = BoolProperty(name = "Add Edges", default = True, description = "Add Edges to the resulting Mesh [not used when Add Faces is checked]")
    
    MinAngle = FloatProperty(name = "Min Angle", default = math.radians(0.0), soft_min = math.radians(-180.0), soft_max = math.radians(180.0), unit = 'ROTATION', description = "Minimum angle (where offset reaches its maximum)")
    MaxAngle = FloatProperty(name = "Max Angle", default = math.radians(180.0), soft_min = math.radians(-180.0), soft_max = math.radians(180.0), unit = 'ROTATION', description = "Maximum angle (where offset reaches its minimum)")
    
    MinOffset = FloatProperty(name = "Min Offset", soft_min = 0.0, default = 0.0, description = "Minimum Offset")
    MaxOffset = FloatProperty(name = "Max Offset", soft_min = 0.0, default = 0.2, description = "Maximum Offset")
    ScaleOffset = FloatProperty(name = "Scale Offset", soft_min = 0.0, default = 1.0, description = "(Global) scale factor for calculated Offset")
    BiasOffset = FloatProperty(name = "Bias Offset", soft_min = 0.0, soft_max = 1.0, default = 0.5, description = "Bias for Offset direction [0.5: bisection of edge directions]")
    
    OffsetIn = FloatProperty(name = "Offset In", soft_min = 0.0, soft_max = 1.0, default = 0.5, description = "Inward offset for the guide used to construct the outer vertices")
    OffsetInBias = FloatProperty(name = "Offset In Bias", soft_min = 0.0, soft_max = 1.0, default = 0.5, description = "Bias for inward offset direction [0.5: midpoint of offset vertices]")
    
    VertSlideBias1 = FloatProperty(name = "Vert Slide Bias1", soft_min = 0.0, soft_max = 1.0, default = 0.2, description = "Bias for first final outer vertex")
    VertSlideBias2 = FloatProperty(name = "Vert Slide Bias2", soft_min = 0.0, soft_max = 1.0, default = 0.2, description = "Bias for second final outer vertex")
    
    VertProjectFactor1 = FloatProperty(name = "Vert Project Factor1", soft_min = 0.0, soft_max = 1.0, default = 0.5, description = "Projection factor/strength for first final outer vertex")
    VertProjectFactor2 = FloatProperty(name = "Vert Project Factor2", soft_min = 0.0, soft_max = 1.0, default = 0.5, description = "Projection factor/strength for second final outer vertex")
    
    @classmethod
    def poll(self, context):
        return Smoother.Smoother.poll()
            
    def draw(self, context):
        layout = self.layout.column(align = True)
        
        split = layout.split()
        loopLayout = split.column(align = True)
        loopLayout.label(text = "Loop:")
        loopLayout.prop(self, 'InvertLoop', text = "Invert")
        
        geoLayout = split.column(align = True)
        geoLayout.label(text = "Geometry:")
        geoLayout.prop(self, 'AddFaces')
        if not self.AddFaces:
            geoLayout.prop(self, 'AddEdges')
        
        split = layout.split()
        angLayout = split.column(align = True)
        angLayout.label(text = "Angle:")
        angLayout.prop(self, 'MinAngle', text = "Min")
        angLayout.prop(self, 'MaxAngle', text = "Max")
        
        offLayout = split.column(align = True)
        offLayout.label(text = "Offset:")
        offLayout.prop(self, 'MinOffset', text = "Min")
        offLayout.prop(self, 'MaxOffset', text = "Max")
        offLayout.prop(self, 'ScaleOffset', text = "Scale")
        offLayout.prop(self, 'BiasOffset', text = "Bias")
        
        split = layout.split()
        inwLayout = split.column(align = True)
        inwLayout.label(text = "Guide construction:")
        inwLayout.prop(self, 'OffsetIn', text = "Factor")
        inwLayout.prop(self, 'OffsetInBias', text = "Bias")
        
        vsLayout = split.column(align = True)
        vsLayout.label(text = "Outer Verts Slide:")
        vsLayout.prop(self, 'VertSlideBias1', text = "Bias1")
        vsLayout.prop(self, 'VertSlideBias2', text = "Bias2")
        
        split = layout.split()
        vpLayout = split.column(align = True)
        vpLayout.label(text = "Inner Verts Project:")
        vpLayout.prop(self, 'VertProjectFactor1', text = "Factor1")
        vpLayout.prop(self, 'VertProjectFactor2', text = "Factor2")
        
    def execute(self, context):
        smoother = Smoother.Smoother(self, context)
        smoother.AddMesh()
            
        return {'FINISHED'}


        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        