import bpy
import bmesh
import math

from . import Util
from . import Math

class Smoother:
    @classmethod
    def poll(self):
        if not bpy.context.mode == 'EDIT_MESH': return False
        
        actObj = bpy.context.active_object
        if actObj is None: return False         # ? is this really necessary when in 'EDIT_MESH' mode?
        if actObj.type != 'MESH': return False  # ? is this really necessary when in 'EDIT_MESH' mode?
        
        bm = bmesh.from_edit_mesh(actObj.data)        
        selEdges = Util.getSelectedEdges(bm)
        nrSelEdges = len(selEdges)
        if nrSelEdges < 3: return False
        
        return True
        
    def __init__(self, operator, context):
        self.operator = operator
        self.context = context
        
        self.actObj = bpy.context.active_object
        self.bm = bmesh.from_edit_mesh(self.actObj.data)
        self.selEdges = Util.getSelectedEdges(self.bm)
        self.nrSelEdges = len(self.selEdges)
        
        self.SetupLoopVerts()
        self.SetupLoopVertsData()
        self.SetupOffsets(self.operator)
        
        
    def SetupLoopVerts(self):
        # 1st edge
        currEdge = self.selEdges[0]
        firstVert = currEdge.verts[0]
        lastVert = currEdge.verts[1]
        self.loopVerts = [firstVert, lastVert]
        
        # 2nd edge
        usedEdges = []
        for nextEdge in self.selEdges[1:]:
            if Util.edgeContainsVertex(nextEdge, lastVert):
                self.loopVerts.append(nextEdge.other_vert(lastVert))
                lastVert = self.loopVerts[-1]
                usedEdges.append(nextEdge)
                break
                
        if len(usedEdges) < 1:
            print("??", "len(usedEdges) < 1")
        
        # 3rd++ edges
        for iVert in range(self.nrSelEdges - 3):
            for nextEdge in self.selEdges:
                if nextEdge in usedEdges: continue
                if Util.edgeContainsVertex(nextEdge, lastVert):
                    self.loopVerts.append(nextEdge.other_vert(lastVert))
                    lastVert = self.loopVerts[-1]
                    usedEdges.append(nextEdge)
                    break
                    
            if len(usedEdges) < (2 + iVert):
                print("??", "len(usedEdges) < (2 + iVert)")
        
        if self.operator.InvertLoop:
            self.loopVerts = list(reversed(self.loopVerts))
        
        self.nrLoopVerts = len(self.loopVerts)
                
    def SetupLoopVertsData(self):
        self.lvDirsOffset  = []
        self.lvAnglesIn    = []
        for ivCurr in range(self.nrLoopVerts):
            ivPrev = ivCurr - 1
            ivNext = ivCurr + 1
            if ivNext >= self.nrLoopVerts: ivNext -= self.nrLoopVerts
            
            dirPrev = (self.loopVerts[ivCurr].co - self.loopVerts[ivPrev].co).normalized()
            dirNext = (self.loopVerts[ivNext].co - self.loopVerts[ivCurr].co).normalized()
            
            bias = self.operator.BiasOffset
            dirOffset = (dirPrev * bias + dirNext * (bias - 1.0)).normalized()
            self.lvDirsOffset.append(dirOffset)
            self.lvAnglesIn.append(dirNext.angle(-dirPrev, math.pi))
            
        self.SetupOffsets(self.operator)
        
        self.oDirsTangent = []
        # self.lvNormals     = []   # not used
        for ivCurr in range(self.nrLoopVerts):
            ivPrev = ivCurr - 1
            ivNext = ivCurr + 1
            if ivNext >= self.nrLoopVerts: ivNext -= self.nrLoopVerts
            
            dirPrev = (self.oPositions[ivCurr] - self.oPositions[ivPrev]).normalized()
            dirNext = (self.oPositions[ivNext] - self.oPositions[ivCurr]).normalized()
            self.oDirsTangent.append(Util.getBisection(dirPrev, dirNext))
        
        self.SetupOffsets2()
        
        self.SetupOuterVerts()
     
    #  funcData == operator -- or not..
    def SetupOffsets(self, funcData):
        self.SetupOffsetCalculation(funcData)
        
        self.lvOffsets = []
        self.oPositions = []
        for iLV in range(self.nrLoopVerts):
            offset = self.CalculateOffset(self.lvAnglesIn[iLV])
            self.lvOffsets.append(offset)
            offPos = self.loopVerts[iLV].co + (self.lvDirsOffset[iLV] * offset)
            self.oPositions.append(offPos)
        
    #  funcData == operator -- or not..
    def SetupOffsetCalculation(self, funcData):
        self.slope = (funcData.MinOffset - funcData.MaxOffset) / (funcData.MaxAngle - funcData.MinAngle)
        self.val0 = funcData.MaxOffset - (funcData.MinAngle * self.slope)
        self.scale = funcData.ScaleOffset
        
    def CalculateOffset(self, angle):
        return (self.val0 + (angle * self.slope)) * self.scale
        
    def SetupOffsets2(self):
        self.oPositions2 = []
        for ivCurr in range(self.nrLoopVerts):
            ivNext = ivCurr + 1
            if ivNext >= self.nrLoopVerts: ivNext -= self.nrLoopVerts
            
            coPos = self.oPositions[ivCurr]
            noPos = self.oPositions[ivNext]
            line1 = Math.Line(coPos, self.oDirsTangent[ivCurr])
            line2 = Math.Line(noPos, - self.oDirsTangent[ivNext])
            par1, par2 = line1.CalcParsClosestToLine(line2)
            v1 = line1.CalcPoint(par1)
            v2 = line2.CalcPoint(par2)
            v0 = (v1 + v2) * 0.5
            
            # inward offset
            bias = self.operator.OffsetInBias
            target = coPos * (1.0 - bias) + noPos * bias
            dir = target - v0
            v = v0 + dir * self.operator.OffsetIn
            
            self.oPositions2.append(v)

    def SetupOuterVerts(self):
        self.ovPositions = []   # list of lists[2] of vertex positions
        self.ovPositions2 = []  # list of lists[2] of vertex positions
        for ivCurr in range(self.nrLoopVerts):
            ivNext = ivCurr + 1
            if ivNext >= self.nrLoopVerts: ivNext -= self.nrLoopVerts

            # outer
            coPos = self.oPositions[ivCurr]
            noPos = self.oPositions[ivNext]
            moPos = self.oPositions2[ivCurr]
            
            bias1 = self.operator.VertSlideBias1
            v1 = coPos * (1.0 - bias1) + moPos * bias1
            bias2 = self.operator.VertSlideBias2
            v2 = noPos * (1.0 - bias2) + moPos * bias2
            
            self.ovPositions.append([v1, v2])
            
            # inner/2
            len1 = (v1 - coPos).magnitude
            len2 = (v2 - v1).magnitude
            len3 = (noPos - v2).magnitude
            lentot = len1 + len2 + len3
            fac1 = len1 / lentot
            fac2 = (len1 + len2) / lentot
            
            clvPos = self.loopVerts[ivCurr].co
            nlvPos = self.loopVerts[ivNext].co
            lvDir = nlvPos - clvPos
            target1 = clvPos + lvDir * fac1
            target2 = clvPos + lvDir * fac2
            
            pfac1 = self.operator.VertProjectFactor1
            tarDir1 = target1 - v1
            v21 = v1 + tarDir1 * pfac1
            pfac2 = self.operator.VertProjectFactor2
            tarDir2 = target2 - v2
            v22 = v2 + tarDir2 * pfac2
    
            self.ovPositions2.append([v21, v22])
    
    
    # verts/loopvert: 6, in edgeloop order, inward to outward
    def GetVerts(self):
        rvlV = []
        
        for ivCurr in range(self.nrLoopVerts):
            rvlV.append(self.loopVerts[ivCurr].co)  # original edgeloop vert
            rvlV.append(self.oPositions[ivCurr])    # offset vert
            
            ovl1 = self.ovPositions[ivCurr]
            ovl2 = self.ovPositions2[ivCurr]
        
            rvlV.append(ovl2[0])                    # inner outer vert 1
            rvlV.append(ovl1[0])                    # outer outer vert 1
            rvlV.append(ovl2[1])                    # inner outer vert 2
            rvlV.append(ovl1[1])                    # outer outer vert 2
        
        return rvlV
        
    def GetEdges(self):
        rvlE = []
        
        nrVerts = self.nrLoopVerts * 6
        for ivCurr in range(self.nrLoopVerts):
            ivOrig = ivCurr * 6
            ivOffset = ivOrig + 1
            
            ivIO1 = ivOffset + 1
            ivOO1 = ivIO1 + 1
            ivIO2 = ivOO1 + 1
            ivOO2 = ivIO2 + 1
            
            ivNextOrig = ivOO2 + 1
            if ivNextOrig >= nrVerts: ivNextOrig -= nrVerts
            ivNextOffset = ivNextOrig + 1
            
            # 'perpendicular'/radial edges
            rvlE.append([ivOrig, ivOffset])
            rvlE.append([ivIO1, ivOO1])
            rvlE.append([ivIO2, ivOO2])
            
            # 'parallel' edges: inner
            rvlE.append([ivOrig, ivIO1])
            rvlE.append([ivIO1, ivIO2])
            rvlE.append([ivIO2, ivNextOrig])
            
            # 'parallel' edges: outer
            rvlE.append([ivOffset, ivOO1])
            rvlE.append([ivOO1, ivOO2])
            rvlE.append([ivOO2, ivNextOffset])
        
        return rvlE
        
    def GetFaces(self):
        rvlF = []
        
        nrVerts = self.nrLoopVerts * 6
        for ivCurr in range(self.nrLoopVerts):
            ivOrig = ivCurr * 6
            ivOffset = ivOrig + 1
            
            ivIO1 = ivOffset + 1
            ivOO1 = ivIO1 + 1
            ivIO2 = ivOO1 + 1
            ivOO2 = ivIO2 + 1
            
            ivNextOrig = ivOO2 + 1
            if ivNextOrig >= nrVerts: ivNextOrig -= nrVerts
            ivNextOffset = ivNextOrig + 1
            
            rvlF.append([ivOrig, ivIO1, ivIO2, ivNextOrig])         # inner
            rvlF.append([ivOrig, ivOffset, ivOO1, ivIO1])           # outer 1
            rvlF.append([ivIO1, ivOO1, ivOO2, ivIO2])               # outer 2
            rvlF.append([ivIO2, ivOO2, ivNextOffset, ivNextOrig])   # outer 3
        
        return rvlF

    def AddMesh(self):
        rvlV = []
        rvlE = []
        rvlF = []
        
        rvlV = self.GetVerts()
        
        if self.operator.AddFaces:
            rvlE = []
            rvlF = self.GetFaces()
        else:
            rvlF = []
            if self.operator.AddEdges:
                rvlE = self.GetEdges()
            else: rvlE = []
        
        # that cursor thing again..
        origCursorLoc = bpy.context.scene.cursor_location.copy()
        bpy.context.scene.cursor_location = self.actObj.location
        newObj = Util.addMesh(self.operator, self.context, "Smoother", rvlV, rvlE, rvlF)
        bpy.context.scene.cursor_location = origCursorLoc
        
        
        
    def Log(self):
        print("Smoother" + "." + "Log" + "()")
        
        print("--", "actObj.name: %s" % self.actObj.name)
        print("--", "nrSelEdges:  %d" % self.nrSelEdges)
        
    def LogVerts(self):
        print("Smoother" + "." + "LogVerts" + "()")

        rvlV = self.GetVerts()
        nrVerts = len(rvlV)
        
        for ivCurr in range(self.nrLoopVerts):
            print("--", "vert: %d" % (ivCurr + 1))
            
            ivOrig = ivCurr * 6
            ivOffset = ivOrig + 1
            
            ivIO1 = ivOffset + 1
            ivOO1 = ivIO1 + 1
            ivIO2 = ivOO1 + 1
            ivOO2 = ivIO2 + 1
            
            ivNextOrig = ivOO2 + 1
            if ivNextOrig >= nrVerts: ivNextOrig -= nrVerts
            ivNextOffset = ivNextOrig + 1

            print("--", "--", "Orig:   ", rvlV[ivOrig])
            print("--", "--", "Offset: ", rvlV[ivOffset])
            print("--", "--", "oPos2:  ", self.oPositions2[ivCurr])
            print("--", "--", "oDirTan:", self.oDirsTangent[ivCurr])
            
            lOO = self.ovPositions[ivCurr]
            print("--", "--", "OOuter1:", lOO[0])
            print("--", "--", "OOuter2:", lOO[1])
            lIO = self.ovPositions2[ivCurr]
            print("--", "--", "IOuter1:", lIO[0])
            print("--", "--", "IOuter2:", lIO[1])
        
    def LogSelEdges(self):
        print("Smoother" + "." + "LogSelEdges" + "(): %d" % self.nrSelEdges)
        
        for bmEdge in self.selEdges:
            print("--", "%03d, %03d" % (bmEdge.verts[0].index, bmEdge.verts[1].index))
        
    def LogVertLoop(self):
        print("Smoother" + "." + "LogVertLoop" + "(): %d" % self.nrLoopVerts)
        
        for bmVert in self.loopVerts:
            print("--", "%03d" % bmVert.index)
        
    def LogAngles(self):
        nrAngles = len(self.lvAnglesIn)
        print("Smoother" + "." + "LogAngles" + "(): %d" % nrAngles)
        
        for ang in self.lvAnglesIn:
            print("--", "%1.3f" % ang)
        
    def LogOffsets(self):
        nrOffsets = len(self.lvOffsets)
        print("Smoother" + "." + "LogOffsets" + "(): %d" % nrOffsets)
        
        for off in self.lvOffsets:
            print("--", "%1.3f" % off)
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        
        